#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "common.h"
#include "macros.h"
#include "buffer.h"
#include "failures.h"
#include "misc.h"
#include "hal-proto.h"
#include "hal-cmd.h"

#define SIDE_MASTER	"M"
#define SIDE_SLAVE	"S"

static void hal_cmd_build_request(struct buffer *pkt, const char *cmd,
				  const char *payload, int plen)
{
	buffer_printf_append(pkt, "%s\n%s\n%s\n%08X\n\n%s",
				HAL_PROTO_MAGIC, cmd, HAL_SUBCMD_REQUEST, plen,
				payload);
}

static void hal_cmd_build_reply(struct buffer *pkt, const char *cmd,
				  const char *payload, int plen)
{
	buffer_printf_append(pkt, "%s\n%s\n%s\n%08X\n\n%s",
				HAL_PROTO_MAGIC, cmd, HAL_SUBCMD_REPLY, plen,
				payload);
}

/* Build request commands. */
void hal_cmd_connect_request(struct buffer *pkt)
{
	char pbuf[HAL_PAYLOAD_MAX+1];
	int res;

	res = xsnprintf(pbuf, sizeof(pbuf), "%08X\n%08X\n",
			HAL_PROTO_MAJOR, HAL_PROTO_MINOR);

	hal_cmd_build_request(pkt, HAL_CMD_CONNECT, pbuf, res);
}

void hal_cmd_connect_reply(struct buffer *pkt, unsigned int inputs_sz,
			   unsigned int outputs_sz, int master)
{
	char pbuf[HAL_PAYLOAD_MAX+1];
	int res;

	res = xsnprintf(pbuf, sizeof(pbuf), "%s\n%08X\n%08X\n%s\n",
			HAL_OK, inputs_sz, outputs_sz,
			master ? SIDE_MASTER : SIDE_SLAVE);

	hal_cmd_build_reply(pkt, HAL_CMD_CONNECT, pbuf, res);
}

static void hal_cmd_noargs_request(struct buffer *pkt, const char *cmd)
{
	/* Payload is empty but we send zero byte. */
	hal_cmd_build_request(pkt, cmd, "00", 2);
}

void hal_cmd_disconnect_request(struct buffer *pkt)
{
	hal_cmd_noargs_request(pkt, HAL_CMD_DISCONNECT);
}

void hal_cmd_cycle_begin_request(struct buffer *pkt)
{
	hal_cmd_noargs_request(pkt, HAL_CMD_CYCLE_BEGIN);
}

void hal_cmd_cycle_run_request(struct buffer *pkt)
{
	hal_cmd_noargs_request(pkt, HAL_CMD_CYCLE_RUN);
}

void hal_cmd_cycle_end_request(struct buffer *pkt)
{
	hal_cmd_noargs_request(pkt, HAL_CMD_CYCLE_END);
}

void hal_cmd_write_prepare_request(struct buffer *pkt)
{
	hal_cmd_noargs_request(pkt, HAL_CMD_WRITE_PREPARE);
}

void hal_cmd_write_end_request(struct buffer *pkt)
{
	hal_cmd_noargs_request(pkt, HAL_CMD_WRITE_END);
}

void hal_cmd_read_prepare_request(struct buffer *pkt)
{
	hal_cmd_noargs_request(pkt, HAL_CMD_READ_PREPARE);
}

void hal_cmd_read_end_request(struct buffer *pkt)
{
	hal_cmd_noargs_request(pkt, HAL_CMD_READ_END);
}

/* Build reply commands. */
static void hal_cmd_noargs_reply(struct buffer *pkt, const char *cmd)
{
	char pbuf[HAL_PAYLOAD_MAX+1];
	int res;

	res = xsnprintf(pbuf, sizeof(pbuf), "%s\n", HAL_OK);

	hal_cmd_build_reply(pkt, cmd, pbuf, res);
}

void hal_cmd_disconnect_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_DISCONNECT);
}

void hal_cmd_cycle_begin_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_CYCLE_BEGIN);
}

void hal_cmd_cycle_run_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_CYCLE_RUN);
}

void hal_cmd_cycle_end_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_CYCLE_END);
}

void hal_cmd_write_inputs_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_WRITE_INPUTS);
}

void hal_cmd_write_outputs_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_WRITE_OUTPUTS);
}

void hal_cmd_write_prepare_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_WRITE_PREPARE);
}

void hal_cmd_write_end_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_WRITE_END);
}

void hal_cmd_read_prepare_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_READ_PREPARE);
}

void hal_cmd_read_end_reply(struct buffer *pkt)
{
	hal_cmd_noargs_reply(pkt, HAL_CMD_READ_END);
}

static void hal_cmd_read_image_request(struct buffer *pkt, const char *cmd,
					unsigned int len, unsigned int off)
{
	char pbuf[HAL_PAYLOAD_MAX+1];
	int res;
	
	res = xsnprintf(pbuf, sizeof(pbuf), "%08X\n%08X\n", len, off);

	hal_cmd_build_request(pkt, cmd, pbuf, res);
}

void hal_cmd_read_inputs_request(struct buffer *pkt, unsigned int len,
				 unsigned int off)
{
	hal_cmd_read_image_request(pkt, HAL_CMD_READ_INPUTS, len, off);
}

void hal_cmd_read_outputs_request(struct buffer *pkt, unsigned int len,
				  unsigned int off)
{
	hal_cmd_read_image_request(pkt, HAL_CMD_READ_OUTPUTS, len, off);
}

static void hal_cmd_write_image_request(struct buffer *pkt, const char *cmd, unsigned int off,
					const char image[HAL_IMAGE_MAX+1])
{
	char pbuf[HAL_PAYLOAD_MAX+1];
	size_t len;
	int res;

	len = strnlen(image, HAL_IMAGE_MAX+1);
	if (len == HAL_IMAGE_MAX+1)
		xerror(1, "%s image size too large", cmd);
	
	res = xsnprintf(pbuf, sizeof(pbuf), "%08X\n%s\n", off, image);

	hal_cmd_build_request(pkt, cmd, pbuf, res);
}

void hal_cmd_write_inputs_request(struct buffer *pkt,  unsigned int off,
				  const char inputs[HAL_IMAGE_MAX+1])
{
	hal_cmd_write_image_request(pkt, HAL_CMD_WRITE_INPUTS, off, inputs);
}

void hal_cmd_write_outputs_request(struct buffer *pkt, unsigned int off,
				   const char outputs[HAL_IMAGE_MAX+1])
{
	hal_cmd_write_image_request(pkt, HAL_CMD_WRITE_OUTPUTS, off, outputs);
}

static void hal_cmd_read_image_reply(struct buffer *pkt, const char *cmd,
				     unsigned int off, const char image[HAL_IMAGE_MAX+1])
{
	char pbuf[HAL_PAYLOAD_MAX+1];
	size_t len;
	int res;

	len = strnlen(image, HAL_IMAGE_MAX+1);
	if (len == HAL_IMAGE_MAX+1)
		xerror(1, "%s image size too large", cmd);
	
	res = xsnprintf(pbuf, sizeof(pbuf), "%s\n%08X\n%s\n",
				HAL_OK, off, image);

	hal_cmd_build_reply(pkt, cmd, pbuf, res);
}

void hal_cmd_read_inputs_reply(struct buffer *pkt, unsigned int off,
			       const char inputs[HAL_IMAGE_MAX+1])
{
	hal_cmd_read_image_reply(pkt, HAL_CMD_READ_INPUTS, off, inputs);
}

void hal_cmd_read_outputs_reply(struct buffer *pkt, unsigned int off,
				const char outputs[HAL_IMAGE_MAX+1])
{
	hal_cmd_read_image_reply(pkt, HAL_CMD_READ_OUTPUTS, off, outputs);
}

void hal_status_error_reply(struct buffer *pkt, const char *cmd, const char *error)
{
	char pbuf[HAL_PAYLOAD_MAX+1];
	int res;

	res = xsnprintf(pbuf, sizeof(pbuf), "%s\n%s\n", HAL_ERR, error);

	hal_cmd_build_reply(pkt, cmd, pbuf, res);
}

